module Chromate

  # 
  # The `Effect` class is used to represent a generic escape sequences,
  # providing a standardized escape mechanism. You may convert an effect to
  # either a string, in which case it will be escaped as an SGR escape sequence,
  # or an integer, in which case it will yield its internal value. You may also
  # explicitly escape an effect by calling {#escape} on it.
  # 
  class Effect

    # 
    # Create a new effect with the specified value.
    # @param  value [Fixnum]
    # 
    # @return [Effect]
    def initialize(value)
      unless value.is_a?(Numeric)
        raise TypeError, "cannot create #{self.class} from #{value.class}"
      end

      @value = value.to_i
    end

    # 
    # Get the effect with the specified name.
    # @param  value [String,Symbol] the name of an effect
    # 
    # @return [Effect]
    def self.[](value)
      if value.is_a?(Symbol) or value.is_a?(String)
        EFFECTS[value.intern]
      else
        raise TypeError, "don't know what to do with a #{value.class}"
      end
    end

    # 
    # Escape as an SGR escape sequence.
    # 
    # @return [String] an escape sequence of the form "\e[1;2;3;...m"
    def escape
      "\e[#{@value}m"
    end

    # 
    # Get the name of this effect, if it exists.
    # 
    # @return [Symbol]
    def name
      if EFFECTS.value?(self)
        EFFECTS.find { |k, v| v == self }.first
      end
    end

    # @return [String]
    def inspect
      if name.nil?
        "#<Effect: #{@value}>"
      else
        "#<Effect: #{name}>"
      end
    end

    # 
    # Yield the internal value.
    # 
    # @return [Integer]
    def to_i
      @value
    end

    # @return [String]
    def to_s
      escape
    end

    # @return [Boolean]
    def ==(other)
      to_i == other.to_i
    end

    def +(other)
      case other
      when EffectSet
        other.add(self)
      when Effect
        EffectSet.new([self, other])
      when String
        escape + other
      else
        raise TypeError, "cannot add #{other.class} to #{self.class}"
      end
    end

  end

end